/* 
 * Copyright 2015-2016 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SPOOFER_MANAGER_COMMON_COMMON_H
#define SPOOFER_MANAGER_COMMON_COMMON_H

#include "port.h"
#include <typeinfo>

#define QT_NO_CAST_FROM_ASCII
#include <QString>
#include <QStringBuilder> // efficient string concatenation with operator%
#define QSL(str) QStringLiteral(str)

#include <QDataStream>
#include <QCoreApplication>
#include <QFile>
#include <QHostAddress>
#include <QDateTime>
#include <QSettings>
#include <QList>

#ifdef __GNUC__
# define NORETURN                 __attribute__((__noreturn__)) void
# define UNUSED                   __attribute__((__unused__))
# define USED                     __attribute__((__used__))
# define FORMAT_PRINTF(fmt, var)  __attribute__((__format__(printf, fmt, var)))
#else
# define NORETURN                 void
# define UNUSED                   /* empty */
# define USED                     /* empty */
# define FORMAT_PRINTF(fmt, var)  /* empty */
#endif

// forward declarations
QT_BEGIN_NAMESPACE
class QCommandLineParser;
QT_END_NAMESPACE

#define EVERYONE_IS_PRIVILEGED 1

// scheduler messages
enum sc_msg_type {
    SC_ERROR,
    SC_TEXT,
    SC_SCHEDULED,
    SC_PROBER_STARTED,
    SC_PROBER_FINISHED,
    SC_PROBER_ERROR,
    SC_PAUSED,
    SC_RESUMED,
    SC_DONE_CMD,
    SC_CONFIG_CHANGED
};

struct sc_msg_text {
    QString text;
};

inline QDataStream &operator<<(QDataStream &out, const sc_msg_text &msg) {
    return out << msg.text;
}

inline QDataStream &operator>>(QDataStream &in, sc_msg_text &msg) {
    return in >> msg.text;
}

struct sc_msg_scheduled {
    // QHostAddress addr;
    qint32 when;
    sc_msg_scheduled() : /*addr(),*/ when(0) {}
};

inline QDataStream &operator<<(QDataStream &out, const sc_msg_scheduled &msg) {
    return out /*<< msg.addr*/ << msg.when;
}

inline QDataStream &operator>>(QDataStream &in, sc_msg_scheduled &msg) {
    return in /*>> msg.addr*/ >> msg.when;
}

#if 0
struct sc_msg_started {
    QHostAddress addr;
    QString logfile;
};

inline QDataStream &operator<<(QDataStream &out, const sc_msg_started &msg) {
    return out << msg.addr << msg.logfile;
}

inline QDataStream &operator>>(QDataStream &in, sc_msg_started &msg) {
    return in >> msg.addr >> msg.logfile;
}

struct sc_msg_success {
    QHostAddress addr;
    QString logfile;
    QString url;
};

inline QDataStream &operator<<(QDataStream &out, const sc_msg_success &msg) {
    return out << msg.addr << msg.logfile << msg.url;
}

inline QDataStream &operator>>(QDataStream &in, sc_msg_success &msg) {
    return in >> msg.addr >> msg.logfile >> msg.url;
}

struct sc_msg_failed {
    QHostAddress addr;
    QString logfile;
};

inline QDataStream &operator<<(QDataStream &out, const sc_msg_failed &msg) {
    return out << msg.addr << msg.logfile;
}

inline QDataStream &operator>>(QDataStream &in, sc_msg_failed &msg) {
    return in >> msg.addr >> msg.logfile;
}

struct sc_msg_disabled {
    QHostAddress addr;
    QDateTime when;
};

inline QDataStream &operator<<(QDataStream &out, const sc_msg_disabled &msg) {
    return out << msg.addr << msg.when;
}

inline QDataStream &operator>>(QDataStream &in, sc_msg_disabled &msg) {
    return in >> msg.addr >> msg.when;
}

struct sc_msg_disconnected {
    QHostAddress addr;
};

inline QDataStream &operator<<(QDataStream &out, const sc_msg_disconnected &msg) {
    return out << msg.addr;
}

inline QDataStream &operator>>(QDataStream &in, sc_msg_disconnected &msg) {
    return in >> msg.addr;
}
#endif

// base class for Spoofer applications
class SpooferBase {
public:
    // A write-only QIODevice that wraps another but doesn't open it until
    // needed, and can be switched to a different device at any time.
    class OnDemandDevice : public QIODevice {
    private:
	QIODevice *dev, *newdev, *fallback;
	QString newname;
    public:
	OnDemandDevice(FILE *file) : QIODevice(), newdev(), newname() {
	    QFile *qfile = new QFile();
	    if (qfile) qfile->open(file, WriteOnly|Unbuffered);
	    dev = qfile;
	    this->open(WriteOnly);
	}
	qint64 readData(char *data, qint64 maxSize)
	    { Q_UNUSED(data); Q_UNUSED(maxSize); return -1; }
	qint64 writeData(const char *data, qint64 maxSize);
	void setDevice(QIODevice *device, const QString &name) {
	    newname = name;
	    newdev = device;
	}
	void setDevice(QFileDevice *device) {
	    setDevice(device, device->fileName());
	}
	void setFallbackDevice(QFileDevice *device) {
	    fallback = device;
	}
	void close() {
	    QIODevice::close();
	    if (dev) { delete dev; dev = nullptr; }
	    if (newdev) { delete newdev; newdev = nullptr; }
	    if (fallback) { delete fallback; fallback = nullptr; }
	}
	const std::type_info& type() { return dev ? typeid(*dev) : typeid(0); }
    };

    class Config {
    public:
	struct MemberBase {
	    const QString key;
	    const QVariant defaultVal;
	    const QString desc;
	    MemberBase(QString _key, QVariant _defaultVal, QString _desc = QString()) :
		key(_key), defaultVal(_defaultVal), desc(_desc)
	    {
		if (!desc.isNull())
		    members.push_back(this);
	    }
	    virtual ~MemberBase() {}
	    virtual QVariant variant() const = 0;
	    virtual bool setFromString(QString value) = 0;
	    void remove()            // remove: config->foo.remove()
		{ settings->remove(key); }
	};
	template <class T> struct Member : public MemberBase {
	    Member(QString _key = QString(), T _defaultVal = T(), QString _desc = QString()) :
		MemberBase(_key, QVariant(_defaultVal), _desc) {}
	    ~Member() {}
	    T operator()() const // get: config->foo()
		{ return settings->value(key, defaultVal).template value<T>(); }
	    QVariant variant() const
		{ return settings->value(key, defaultVal); }
	    void operator()(T value) // set: config->foo(value)
		{ settings->setValue(key, QVariant(value)); }
	    bool setFromString(QString value) {
		QVariant var(value);
		int qtypeid = qMetaTypeId<T>();
		if (!var.convert(qtypeid)) {
		    qDebug() << "Failed to convert" << value << "to" << 
			QMetaType::typeName(qtypeid);
		    return false;
		}
		settings->setValue(key, var);
		return true;
	    }
	};

	static QSettings *settings;
    public:
	static QList<MemberBase*> members;

	Member<QString> dataDir;
	Member<QString> schedulerSocketName;
	Member<bool> enableIPv4;
	Member<bool> enableIPv6;
	Member<int> keepLogs;
	Member<bool> shareData;
	Member<int> proberInterval;
	Member<int> proberRetryInterval;
	Member<int> delayInterval;
	Member<int> netPollInterval;
	Member<bool> unprivView;
	Member<bool> unprivTest;
	Member<bool> unprivPref;

	Config(bool debug) :
	    // Internal
	    dataDir(QSL("dataDir")),
	    schedulerSocketName(QSL("schedulerSocketName")),
	    // General
	    enableIPv4(QSL("enableIPv4"), true,
		QSL("Enable testing on IPv4 interfaces (if available)")),
	    enableIPv6(QSL("enableIPv6"), true,
		QSL("Enable testing on IPv6 interfaces (if available)")),
	    keepLogs(QSL("keepLogs"), 60,
		QSL("Number of prober log files to keep")),
	    shareData(QSL("shareData"), false,
		QSL("Allow test results to be shared publicly on the Spoofer website")),
	    // Intervals:  "Wait this many seconds to..."
	    // odd proberInterval helps prevent many clients from synchronizing
	    proberInterval(QSL("proberInterval"), 7*24*60*60 + 65*60,
		QSL("Run a test after a successful run on the same network")),
	    proberRetryInterval(QSL("proberRetryInterval"), 60,
		QSL("Run a test after the first failed run on the same network")),
	    delayInterval(QSL("delayInterval"), 60,
		QSL("Run a test after scheduler starts or detects a network change")),
	    netPollInterval(QSL("netPollInterval"), 5*60,
		QSL("Check for a network change")),
	    // Permissions:  "Allow unprivileged users on this computer to..."
	    unprivView(QSL("unprivView"), true,
		QSL("Observe a test in progress and view results of past tests")),
	    unprivTest(QSL("unprivTest"), false,
		QSL("Run a test")),
	    unprivPref(QSL("unprivPref"), false,
		QSL("Change preferences"))
	{
	    if (!optSettings.isEmpty()) {
		settings = new QSettings(optSettings, QSettings::IniFormat);
	    } else {
		QSettings::Scope scope = debug ? QSettings::UserScope :
		    QSettings::SystemScope;
		// Mac expects a domain here where others expect a name.
		// QSettings() with no parameters would correctly auto-pick
		// the domain or name, but doesn't let us specify the scope.
		settings = new QSettings(scope,
#ifdef Q_OS_MAC
		    QCoreApplication::organizationDomain(),
#else
		    QCoreApplication::organizationName(),
#endif
		    QCoreApplication::applicationName(), nullptr);
		settings->setFallbacksEnabled(false);
	    }
	    qDebug() << "Config:" << settings->fileName();
	}

	~Config() { delete settings; settings = nullptr; }

	bool error(const char *label) {
	    switch (settings->status()) {
		case QSettings::NoError:
		    qDebug().nospace() << label << "ok";
		    return false;
		case QSettings::AccessError:
		    sperr << label << "AccessError";
		    break;
		case QSettings::FormatError:
		    sperr << label << "FormatError";
		    break;
	    }
	    sperr << " in " << settings->fileName() << endl;
	    return true;
	}

	void sync() {
	    settings->sync();
	    if (error("Config sync: ")) return;
	}

	MemberBase *find(const QString &key) {
	    for (int i = 0; i < members.size(); i++) {
		if (key.compare(members[i]->key) == 0)
		    return members[i];
	    }
	    return nullptr;
	}
    };

    QString appDir;
    QString appFile;
    static bool optDebug;
    static QString optSettings;
    static Config *config;
    static OnDemandDevice outdev;
    static OnDemandDevice errdev;
    static QTextStream spout;
    static QTextStream sperr;
    static const QString proberLogFtime;
    static const QString proberLogGlob;
    static const QString proberLogRegex;

    SpooferBase();

    ~SpooferBase() {
	if (config) delete config;
    }

    static void logHandler(QtMsgType type,
	const QMessageLogContext &ctx, const QString &msg);
    static bool parseCommandLine(QCommandLineParser &clp, QString desc);
    static QString ftime(QString fmt = QString(), const time_t *tp = nullptr);
    static QString ftime_utc(QString fmt = QString(), const time_t *tp = nullptr);

    bool initConfig() {
	config = new Config(optDebug);
	return !config->error("Config: ");
    }
};

#endif // SPOOFER_MANAGER_COMMON_COMMON_H
